-- File: ap_scan_oled_demo.lua
-- Author: RSP @ Embedded Systems Lab (ESL), KMUTNB, BKK/Thailand
-- Date: 2015-04-20
-- Description: 
--   This Lua script for NodeMCU scans the surrounding WiFi APs
--   and displays the found APs one by one on the ssd1306_128x64 I2C OLED.
-- NodeMCU (v0.9.6 build 20150406)
---------------------------------------------------------------------

-------------------------------------------
local sda_pin = 5 -- GPIO14
local scl_pin = 6 -- GPIO12
local oled_addr = 0x3c
i2c.setup(0, sda_pin, scl_pin, i2c.SLOW)

local disp = nil
local disp = u8g.ssd1306_128x64_i2c(oled_addr)

-------------------------------------------
local ap_list = nil
local ap_list_saved = nil
local scan_cnt = 0
local ap_index = 1
-------------------------------------------

function display_ap(num,ssid,mac,chan,rssi)
  disp:firstPage()
  disp:setFont(u8g.font_6x10)
  disp:setFontRefHeightExtendedText()
  disp:setDefaultForegroundColor()
  disp:setFontPosTop()
  
  repeat
    local x=2
    local y=4
    disp:drawRFrame(0, 0, 128-1, 64-1, 1)
    disp:drawStr(x+2, y, 'AP(s) Found: '..num) 
    y = y + 14
    disp:drawStr(x, y, '> '..ssid) 
    y = y + 14
    disp:drawStr(x, y, '> '..mac) 
    y = y + 14
    disp:drawStr(x, y, '> Ch: ' .. chan )
    x = x + 54
    disp:drawStr(x, y, 'S: ' ..rssi..' dBm') 
  until disp:nextPage() == false
end

function scan_ap()
  scan_cnt = (scan_cnt+1) % 5

  if scan_cnt == 1 and ap_list == nil then
    wifi.sta.getap( function(t)
      if t == nil then
         return
      end
      ap_list = {}
      for k,v in pairs(t) do
         table.insert(ap_list, k ..','..v)
      end
    end)
  end

  if ap_list ~= nil then
    if ap_list_saved == nil then
      ap_list_saved = ap_list
      ap_list = nil
    end

    local num = #ap_list_saved
    if ap_index <= num then
       local s = ap_list_saved[ ap_index ]
       local p = '([^,]+),([^,]+),([^,]+),([^,]+),([^,]+)' 
       ssid,sec,rssi,mac,chan = string.match(s,p)
       --print(ssid,mac,chan,rssi)
       display_ap(num,ssid,mac,chan,rssi)
       ap_index = ap_index + 1
    else
       ap_index = 1
       ap_list_saved = nil
    end
  end
end

tmr.alarm( 1, 1000, 1, scan_ap )

